from polytope_feature import shapes

from ..feature import Feature
from ..utils.areas import field_area, get_polygon_area


class Polygons(Feature):
    def __init__(self, feature_config, client_config):
        assert feature_config.pop("type") == "polygon"
        self.shape = feature_config.pop("shape")
        self.max_area = client_config.polygonrules.max_area
        self.area = 0
        self.field_area = 0
        if type(self.shape[0][0]) is not list:
            self.area = get_polygon_area(self.shape)
            if len(self.shape) > client_config.polygonrules.max_points:
                raise ValueError(
                    f"Number of points {len(self.shape)} exceeds the maximum of {client_config.polygonrules.max_points}"  # noqa: E501
                )
            # if self.area > client_config.polygonrules.max_area:
            #    raise ValueError(
            #        f"Area of polygon {self.area} km\u00b2 exceeds the maximum of size of {client_config.polygonrules.max_area} km\u00b2"  # noqa: E501
            #    )
            self.shape = [self.shape]
        else:
            len_polygons = 0
            area_polygons = 0
            for polygon in self.shape:
                len_polygons += len(polygon)
                area_polygons += get_polygon_area(polygon)
            self.area = area_polygons
            if len_polygons > client_config.polygonrules.max_points:
                raise ValueError(
                    f"Number of points {len_polygons} exceeds the maximum of {client_config.polygonrules.max_points}"  # noqa: E501
                )
            # if area_polygons > client_config.polygonrules.max_area:
            #    raise ValueError(
            #        f"Area of polygon {area_polygons} exceeds the maximum of size of {client_config.polygonrules.max_area} degrees\u00b2"  # noqa: E501
            #    )

        if "axes" not in feature_config:
            self.axes = ["latitude", "longitude"]
        else:
            self.axes = feature_config.pop("axes")

        assert len(feature_config) == 0, f"Unexpected keys in config: {feature_config.keys()}"

    def get_shapes(self):
        polygons = []
        for polygon in self.shape:
            points = []
            for point in polygon:
                points.append([point[0], point[1]])
            polygons.append(shapes.Polygon([self.axes[0], self.axes[1]], points))
        return [shapes.Union([self.axes[0], self.axes[1]], *polygons)]

    def incompatible_keys(self):
        return []

    def coverage_type(self):
        return "MultiPoint"

    def name(self):
        return "Polygon"

    def required_keys(self):
        return ["type", "shape"]

    def required_axes(self):
        return ["latitude", "longitude"]

    def parse(self, request, feature_config):
        self.field_area = field_area(request, self.area)
        if "axes" in request:
            if len(request["axes"]) != 2:
                raise ValueError("Polygon feature must have two axes, latitude and longitude")
        # if field_area(request, self.area) > self.max_area:
        #    raise ValueError(
        #        f"The total request size is too large, area of request shape {self.area} * total number of fields = {field_area(request, self.area)} km\u00b2, must be below {self.max_area} km\u00b2 for total size request. "  # noqa: E501
        #    )

        return request



import json

import pytest

from polytope_mars.api import PolytopeMars

# If using a local FDB need to set FDB_HOME and ECCODES_DEFINITIO_PATH


class TestFeatureFactory:
    def setup_method(self):
        self.request = {
            "class": "od",
            "expver": "0001",
            "levtype": "sfc",
            "type": "pf",
            "time": "00:00:00",
            "date": "20231205",
            "step": "0",
            "domain": "g",
            "param": "165/166",
            "stream": "enfo",
            "feature": {
                "type": "frame",
                "outer_box": [[0, 0], [0.2, 0.2]],
                "inner_box": [[0.1, 0.1], [0.15, 0.15]],
            },
            "number": "1",
        }

        self.options = {
            "values": {
                "mapper": {
                    "type": "octahedral",
                    "resolution": 1280,
                    "axes": ["latitude", "longitude"],
                }
            },
            "date": {"merge": {"with": "time", "linkers": ["T", "00"]}},
            "step": {"type_change": "int"},
        }

        self.config = {
            "class": "od",
            "expver": "0001",
            "levtype": "sfc",
            "type": "pf",
        }  # noqa: E501

    @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_frame_invalid(self):
        with pytest.raises(TypeError):
            PolytopeMars(self.options, self.config).extract("invalid")

        with pytest.raises(TypeError):
            PolytopeMars(self.options, self.config).extract({"hello": "world"})  # noqa: E501

        with pytest.raises(TypeError):
            PolytopeMars(self.options, self.config).extract(json.dumps({"hello": "world"}))

        # 'step' is invalid in the request
        self.request["step"] = "0"
        with pytest.raises(TypeError):
            PolytopeMars(self.options, self.config).extract(self.request)

        self.request["levellist"] = "0"
        with pytest.raises(TypeError):
            PolytopeMars(self.options, self.config).extract(self.request)

    @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_frame_valid(self):
        coverage = PolytopeMars(self.config, self.options).extract(self.request)  # noqa: E501
        print(coverage)



from abc import ABC, abstractmethod
from typing import List

from polytope_feature import shapes


class Feature(ABC):
    @abstractmethod
    def name(self):
        pass

    @abstractmethod
    def get_shapes(self) -> List[shapes.Shape]:
        pass

    @abstractmethod
    def incompatible_keys(self) -> List[str]:
        pass

    @abstractmethod
    def coverage_type(self):
        pass

    @abstractmethod
    def parse(self, request, feature_config):
        pass

    @abstractmethod
    def required_keys(self):
        pass

    @abstractmethod
    def required_axes(self):
        pass

    def split_request(self):
        """
        Determines if the request should be split based on the feature configuration.
        This method can be overridden by subclasses to implement specific splitting logic.
        """
        if self.name() == "Polygon" or self.name() == "BoundingBox":
            if self.field_area > self.max_area:
                # If the area of the request exceeds the maximum allowed area, we split the request.
                return True
        # Otherwise, we do not split the request.
        return False

    def validate(self, request, feature_config):
        incompatible_keys = self.incompatible_keys()
        for key in incompatible_keys:
            if key in request:
                raise KeyError(
                    f"Request contains a '{key}' keyword which is not compatible with feature {self.name()}"  # noqa: E501
                )
        required_keys = self.required_keys()
        for key in required_keys:
            if key not in request and key not in feature_config:
                raise KeyError(f"Missing required key {key} not in request")

        required_axes = self.required_axes()
        for key in required_axes:
            if "axes" in feature_config:
                if key not in feature_config["axes"]:
                    raise KeyError(f"Missing required axis {key} not in request")



import logging

from polytope_feature import shapes

from ..feature import Feature
from ..utils.areas import field_area


class Position(Feature):
    def __init__(self, feature_config, client_config):
        assert feature_config.pop("type") == "position"
        self.axes = feature_config.pop("axes", [])

        self.max_size = client_config.polygonrules.max_area

        if self.axes != []:
            if not isinstance(self.axes, list):
                self.axes = ["latitude", "longitude"]
            if self.axes == ["step"] or self.axes == ["date"]:
                raise ValueError(
                    "Position feature cannot have time axis, axes must be latitude and longitude"
                )  # noqa: E501
            for value in self.axes:
                if value not in {"latitude", "longitude"}:
                    raise ValueError("Position feature must have only latitude and longitude axes")  # noqa: E501
        else:
            self.axes = ["latitude", "longitude"]

        self.points = feature_config.pop("points", [])

        assert len(feature_config) == 0, f"Unexpected keys in config: {feature_config.keys()}"

    def get_shapes(self):
        return [
            shapes.Union(
                [self.axes[0], self.axes[1]],
                *[
                    shapes.Point(
                        [self.axes[0], self.axes[1]],
                        [[p[0], p[1]]],
                        method="nearest",  # noqa: E501
                    )
                    for p in self.points
                ],
            ),
        ]

    def incompatible_keys(self):
        return []

    def coverage_type(self):
        return "PointSeries"

    def name(self):
        return "Position"

    def required_keys(self):
        return ["type", "points"]

    def required_axes(self):
        return ["latitude", "longitude"]

    def parse(self, request, feature_config):
        logging.debug("Feature config: %s", feature_config)

        area = field_area(request, len(feature_config["points"]))

        if area > self.max_size:
            raise ValueError(
                f"Number of coordinates*fields for timeseries {area} exceeds total number allowed, please reduce the number of coordinates or fields requested"  # noqa: E501
            )

        if len(feature_config["points"][0]) != 2:
            raise ValueError("Position must have only two values in points")

        logging.debug("After parse request: %s", request)

        return request



from polytope_feature import shapes

from ..feature import Feature


class Frame(Feature):
    def __init__(self, feature_config, client_config):
        assert feature_config.pop("type") == "frame"
        self.points = feature_config.pop("points", [])
        self.outer_box = feature_config.pop("outer_box", [])
        self.inner_box = feature_config.pop("inner_box", [])

        assert len(feature_config) == 0, f"Unexpected keys in config: {feature_config.keys()}"

    def get_shapes(self):
        # frame is a four seperate boxes requested based on the inner and outer boxes  # noqa: E501
        return [
            shapes.Union(
                ["latitude", "longitude"],
                shapes.Box(
                    ["latitude", "longitude"],
                    [self.outer_box[0][0], self.outer_box[0][1]],
                    [self.inner_box[0][0], self.outer_box[1][1]],
                ),
                shapes.Box(
                    ["latitude", "longitude"],
                    [self.inner_box[0][0], self.outer_box[0][1]],
                    [self.inner_box[1][0], self.inner_box[0][1]],
                ),
                shapes.Box(
                    ["latitude", "longitude"],
                    [self.inner_box[0][0], self.inner_box[1][1]],
                    [self.inner_box[1][0], self.outer_box[1][1]],
                ),
                shapes.Box(
                    ["latitude", "longitude"],
                    [self.inner_box[1][0], self.outer_box[0][1]],
                    [self.outer_box[1][0], self.outer_box[1][1]],
                ),
            )
        ]

    def incompatible_keys(self):
        return ["levellist"]

    def coverage_type(self):
        return "MultiPoint"

    def name(self):
        return "Frame"

    def parse(self, request, feature_config):
        return request



from polytope_feature import shapes

from ..feature import Feature
from ..utils.areas import field_area, get_circle_area_from_coords


class Circle(Feature):
    def __init__(self, feature_config, client_config):
        assert feature_config.pop("type") == "circle"
        if len(feature_config["center"][0]) < 2 or len(feature_config["center"][0]) > 3:
            raise ValueError("Circle center must have two values, latitude and longitude")
        self.center = feature_config.pop("center")
        self.max_area = client_config.polygonrules.max_area
        self.radius = feature_config.pop("radius")
        self.area = get_circle_area_from_coords(
            self.center[0][0], self.center[0][1], self.center[0][0] + self.radius, self.center[0][1] + self.radius
        )
        if self.area > client_config.polygonrules.max_area:
            raise ValueError(
                f"Area of circle {self.area} km\u00b2 exceeds the maximum of size of {client_config.polygonrules.max_area} km\u00b2"  # noqa: E501
            )

        if "axes" not in feature_config:
            self.axes = ["latitude", "longitude"]
        else:
            self.axes = feature_config.pop("axes")

        assert len(feature_config) == 0, f"Unexpected keys in config: {feature_config.keys()}"

    def get_shapes(self):
        if len(self.center[0]) == 2:
            return [
                shapes.Disk(
                    [self.axes[0], self.axes[1]],
                    centre=self.center[0],
                    radius=[self.radius, self.radius],
                )
            ]
        else:
            return [
                shapes.Disk(
                    [self.axes[0], self.axes[1], self.axes[2]],
                    centre=self.center[0],
                    radius=[self.radius, self.radius, self.radius],
                )
            ]

    def incompatible_keys(self):
        return []

    def coverage_type(self):
        return "MultiPoint"

    def name(self):
        return "Circle"

    def required_keys(self):
        return ["type", "center", "radius"]

    def required_axes(self):
        return ["latitude", "longitude"]

    def parse(self, request, feature_config):
        if "axes" in feature_config:
            if len(feature_config["center"][0]) != len(feature_config["axes"]):
                raise ValueError("Number of axes must match number of values in center")

        if field_area(request, self.area) > self.max_area:
            raise ValueError(
                "The request size is too large, lower number of fields requested or size of shape requested"  # noqa: E501
            )
        if len(feature_config["center"]) != 1:
            raise ValueError("Circle feature must have one center point")

        return request



import logging

from polytope_feature import shapes

from ..feature import Feature
from ..utils.areas import field_area, get_boundingbox_area


class BoundingBox(Feature):
    def __init__(self, feature_config, client_config):
        assert feature_config.pop("type") == "boundingbox"
        if "points" not in feature_config:
            raise KeyError("Bounding box must have points in feature")
        self.points = feature_config.pop("points", [])
        if "axes" not in feature_config:
            feature_config["axes"] = ["latitude", "longitude"]
        self.axes = feature_config.pop("axes", [])
        self.max_area = client_config.polygonrules.max_area
        self.field_area = 0

        if "axes" in feature_config:
            raise ValueError("Bounding box does not have axes in feature, did you mean axes?")  # noqa: E501

        assert len(feature_config) == 0, f"Unexpected keys in config: {feature_config.keys()}"

        self.area_bb = get_boundingbox_area(self.points)
        logging.info(f"Area of bounding box: {self.area_bb} km\u00b2")

    def get_shapes(self):
        # Time-series is a squashed box from start_step to start_end for each point  # noqa: E501
        if len(self.points[0]) == 2:
            return [
                shapes.Union(
                    ["latitude", "longitude"],
                    *[
                        shapes.Box(
                            ["latitude", "longitude"],
                            lower_corner=[
                                self.points[0][self.axes.index("latitude")],
                                self.points[0][self.axes.index("longitude")],
                            ],  # noqa: E501
                            upper_corner=[
                                self.points[1][self.axes.index("latitude")],
                                self.points[1][self.axes.index("longitude")],
                            ],  # noqa: E501
                        )
                    ],
                )
            ]
        else:
            return [
                shapes.Union(
                    [self.axes[0], self.axes[1], self.axes[2]],
                    *[
                        shapes.Box(
                            [self.axes[0], self.axes[1], self.axes[2]],
                            lower_corner=[
                                self.points[0][0],
                                self.points[0][1],
                                self.points[0][2],
                            ],
                            upper_corner=[
                                self.points[1][0],
                                self.points[1][1],
                                self.points[1][2],
                            ],
                        )
                    ],
                )
            ]

    def incompatible_keys(self):
        return []

    def coverage_type(self):
        return "MultiPoint"

    def name(self):
        return "Bounding Box"

    def required_keys(self):
        return ["type", "points"]

    def required_axes(self):
        return ["latitude", "longitude"]

    def parse(self, request, feature_config):
        if feature_config["type"] != "boundingbox":
            raise ValueError("Feature type must be boundingbox")
        if "axes" in feature_config:
            if len(feature_config["axes"]) < 2 or len(feature_config["axes"]) > 3:
                raise ValueError(
                    "Bounding Box axes must contain 2 or 3 values, latitude, longitude, and optionally levelist"
                )
            if "step" in feature_config["axes"]:
                raise ValueError(
                    "Bounding box axes must be latitude and longitude, step can be requested in main body of request"
                )
            if "latitude" not in feature_config["axes"] or "longitude" not in feature_config["axes"]:
                raise ValueError("Bounding Box axes must contain both latitude and longitude")
            if len(feature_config["axes"]) > 3:
                raise ValueError(
                    "Bounding Box axes must contain at most 3 values, latitude, longitude, and levelist"
                )  # noqa: E501

        self.field_area = field_area(request, self.area_bb)
        # if self.field_area > self.max_area:
        #    raise ValueError(
        #        f"The total request size is too large, area of request shape {self.area_bb} * total number of fields = {field_area(request, self.area_bb)} km\u00b2, must be below {self.max_area} km\u00b2 for total size request. "  # noqa: E501
        #    )

        if len(feature_config["points"]) != 2:
            raise ValueError("Bounding box must have only two points in points")  # noqa: E501
        if "axis" in feature_config:
            raise ValueError("Bounding box does not have axis in feature, did you mean axes?")  # noqa: E501
        if "axes" not in feature_config:
            for point in feature_config["points"]:
                if len(point) != 2:
                    raise ValueError(
                        "For Bounding Box each point must have only two values unless axes is specified"  # noqa: E501
                    )
        else:
            for point in feature_config["points"]:
                if len(point) != len(feature_config["axes"]):
                    raise ValueError("Bounding Box points must have the same number of values as axes")  # noqa: E501
            if "axes" in feature_config:
                if ("levelist" in feature_config["axes"]) and ("levelist" in request):  # noqa: E501
                    raise ValueError("Bounding Box axes is overspecified in request")  # noqa: E501

        return request



from polytope_feature import shapes

from ..feature import Feature


class Path(Feature):
    def __init__(self, feature_config, client_config):
        assert feature_config.pop("type") == "trajectory"
        self.points = feature_config.pop("points", [])
        if "axes" in feature_config:
            self.axes = feature_config.pop("axes")
        else:
            self.axes = ["latitude", "longitude", "levelist", "step"]
        if "inflation" in feature_config:
            if isinstance(feature_config["inflation"], list):
                self.inflation = feature_config.pop("inflation")
                if len(self.inflation) != len(self.axes):
                    raise ValueError(
                        "Inflation must have the same number of values as axes or a single value"  # noqa: E501
                    )
            else:
                self.inflation = []
                infl = feature_config.pop("inflation")
                for ax in self.axes:
                    self.inflation.append(infl)
        if "inflate" in feature_config:
            self.inflate = feature_config.pop("inflate")
        else:
            self.inflate = "round"
        if "axis" in feature_config:
            raise ValueError("Trajectory does not have axis in feature, did you mean axes?")  # noqa: E501

        assert len(feature_config) == 0, f"Unexpected keys in config: {feature_config.keys()}"

    def get_shapes(self):
        if len(self.points[0]) == 2:
            if self.inflate == "round":
                shape = shapes.Disk
            elif self.inflate == "box":
                shape = shapes.Box
            else:
                raise ValueError("Inflate must be either 'round' or 'box' for 2D trajectory feature")  # noqa: E501
            return [
                shapes.Path(
                    [self.axes[0], self.axes[1]],
                    shape(
                        [self.axes[0], self.axes[1]],
                        [0, 0],
                        [self.inflation[0], self.inflation[1]],
                    ),
                    *self.points,
                )
            ]
        elif len(self.points[0]) == 3:
            if self.inflate == "round":
                shape = shapes.Ellipsoid
            elif self.inflate == "box":
                shape = shapes.Box
            else:
                raise ValueError("Inflate must be either 'round' or 'box' for 3D trajectory feature")  # noqa: E501
            return [
                shapes.Path(
                    [self.axes[0], self.axes[1], self.axes[2]],
                    shape(
                        [self.axes[0], self.axes[1], self.axes[2]],
                        [0, 0, 0],
                        [self.inflation[0], self.inflation[1], self.inflation[2]],  # noqa: E501
                    ),
                    *self.points,
                )
            ]
        elif len(self.points[0]) == 4:
            if self.inflate == "round":
                raise ValueError(
                    "Round inflation not yet implemented for 4D trajectory feature, use 'box' instead"  # noqa: E501
                )
            return [
                shapes.Path(
                    [self.axes[0], self.axes[1], self.axes[2], self.axes[3]],
                    shapes.Box(
                        [self.axes[0], self.axes[1], self.axes[2], self.axes[3]],
                        [0, 0, 0, 0],
                        [
                            self.inflation[0],
                            self.inflation[1],
                            self.inflation[2],
                            self.inflation[3],
                        ],  # noqa: E501
                    ),
                    *self.points,
                )
            ]

    def incompatible_keys(self):
        return []

    def valid_axes(self):
        return ["latitude", "longitude", "levelist", "step"]

    def coverage_type(self):
        return "Trajectory"

    def name(self):
        return "Path"

    def required_keys(self):
        return ["type", "points", "inflation"]

    def required_axes(self):
        return ["latitude", "longitude"]

    def parse(self, request, feature_config):
        if "step" in request and "number" in request:
            step = request["step"].split("/")
            number = request["number"].split("/")
            if len(step) > 1 and len(number) > 1:
                raise ValueError("Multiple steps and numbers not yet supported for trajectory feature")  # noqa: E501
        if "step" in request:
            step = request["step"].split("/")
            if len(step) > 1:
                raise ValueError("Multiple steps not yet supported for trajectory feature")
        if "axes" not in feature_config:
            for point in feature_config["points"]:
                if len(point) != 4:
                    raise ValueError(
                        "For Trajectory each point must have only two values unless axes is specified, point must have form ['latitude', 'longitude', 'levelist', 'step']"  # noqa: E501
                    )
        else:
            for point in feature_config["points"]:
                if len(point) != len(feature_config["axes"]):
                    raise ValueError("Trajectory points must have the same number of values as axes")  # noqa: E501
            if ("levelist" in feature_config["axes"]) and ("levelist" in request):  # noqa: E501
                raise ValueError("Trajectory level axes is overspecified in request")  # noqa: E501
        if len(feature_config["points"]) < 2:
            raise ValueError("Trajectory must have atleast two values in points")  # noqa: E501
        if "axes" in feature_config:
            if len(feature_config["axes"]) < 2 or len(feature_config["axes"]) > 4:
                raise ValueError("Trajectory axes must have 2, 3 or 4 values")
            for axis in feature_config["axes"]:
                if axis not in self.valid_axes():
                    raise ValueError(f"Invalid axis: {axis}, must be one of {self.valid_axes()}")  # noqa: E501

        return request



import math

from polytope_mars.utils.areas import request_cost


class TestCosting:
    def setup_method(self):

        self.bbox_request = {
            "class": "od",
            "stream": "enfo",
            "type": "pf",
            "date": "20250114",
            "time": "0000",
            "levtype": "sfc",
            "expver": "0001",
            "domain": "g",
            "param": "167",
            "number": "1",
            "step": "0/to/3",
            "feature": {
                "type": "boundingbox",
                "points": [[0, 0], [1, 1]],
                "axes": ["longitude", "latitude"],
            },
        }

        self.polygon_request = {
            "class": "od",
            "stream": "enfo",
            "type": "pf",
            "date": "20250715",
            "time": "0000",
            "levtype": "sfc",
            "expver": "0001",
            "domain": "g",
            "param": "167",
            "number": "1",
            "step": "0/to/3",
            "feature": {
                "type": "polygon",
                "shape": [[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.0], [0.0, 0.0]],
            },
        }

        self.timeseries_request = {
            "class": "od",
            "stream": "enfo",
            "type": "pf",
            "date": "20250114",
            "time": "0000",
            "levtype": "sfc",
            "expver": "0001",
            "domain": "g",
            "param": "167",
            "number": "1",
            "step": "0/to/3",
            "feature": {
                "type": "timeseries",
                "points": [[-9.10, 38.78]],
                "axes": ["longitude", "latitude"],
            },
        }

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_timeseries_cost(self):
        request_cost_value = request_cost(self.timeseries_request)
        assert request_cost_value == 4

        self.timeseries_request["number"] = "1/to/2"
        assert request_cost(self.timeseries_request) == 8

        self.timeseries_request["step"] = "0/to/6"
        assert request_cost(self.timeseries_request) == 14

        self.timeseries_request["dataset"] = "reanalysis"
        assert request_cost(self.timeseries_request) == 28

    def test_boundingbox_cost(self):
        tolerance = 10

        request_cost_value = request_cost(self.bbox_request)
        assert math.isclose(
            request_cost_value, 49235, abs_tol=tolerance
        ), f"Value {request_cost_value} is not within {tolerance} of {49235}"

        self.bbox_request["number"] = "1/to/2"
        request_cost_value = request_cost(self.bbox_request)
        assert math.isclose(
            request_cost_value, 49235 * 2, abs_tol=tolerance
        ), f"Value {request_cost_value} is not within {tolerance} of {49235 * 2}"

        self.bbox_request["step"] = "0/to/7"
        request_cost_value = request_cost(self.bbox_request)
        assert math.isclose(
            request_cost_value, 49235 * 2 * 2, abs_tol=tolerance
        ), f"Value {request_cost_value} is not within {tolerance} of {49235 * 2 * 2}"

    def test_polygon_cost(self):
        tolerance = 10

        request_cost_value = request_cost(self.polygon_request)
        assert math.isclose(
            request_cost_value, 49235, abs_tol=tolerance
        ), f"Value {request_cost_value} is not within {tolerance} of {49235}"

        self.polygon_request["number"] = "1/to/2"
        request_cost_value = request_cost(self.polygon_request)
        assert math.isclose(
            request_cost_value, 49235 * 2, abs_tol=tolerance
        ), f"Value {request_cost_value} is not within {tolerance} of {49235 * 2}"

        self.polygon_request["step"] = "0/to/7"
        request_cost_value = request_cost(self.polygon_request)
        assert math.isclose(
            request_cost_value, 49235 * 2 * 2, abs_tol=tolerance
        ), f"Value {request_cost_value} is not within {tolerance} of {49235 * 2 * 2}"



import copy
from datetime import datetime, timedelta

import pytest
from conflator import Conflator
from covjsonkit.api import Covjsonkit

from polytope_mars.api import PolytopeMars
from polytope_mars.config import PolytopeMarsConfig

# If using a local FDB need to set GRIBJUMP_CONFIG_FILE and DYLD_LIBRARY_PATH


class TestFeatureFactory:
    def setup_method(self):

        today = datetime.today()
        yesterday = today - timedelta(days=1)
        self.date = yesterday.strftime("%Y%m%d")

        self.request = {
            "class": "od",
            "stream": "enfo",
            "type": "pf",
            "date": self.date,
            "time": "0000",
            "levtype": "pl",
            "expver": "0001",
            "domain": "g",
            "param": "203/133",
            "number": "1",
            "step": "0",
            "levelist": "500",
            "feature": {
                "type": "trajectory",
                "points": [[-1, -1], [0, 0], [1, 1]],
                "axes": ["latitude", "longitude"],
                "inflate": "round",
                "inflation": 0.1,
            },
        }

        self.options = {
            "axis_config": [
                {
                    "axis_name": "date",
                    "transformations": [{"name": "merge", "other_axis": "time", "linkers": ["T", "00"]}],
                },
                {
                    "axis_name": "values",
                    "transformations": [
                        {
                            "name": "mapper",
                            "type": "octahedral",
                            "resolution": 1280,
                            "axes": ["latitude", "longitude"],
                        }
                    ],
                },
                {"axis_name": "latitude", "transformations": [{"name": "reverse", "is_reverse": True}]},
                {"axis_name": "longitude", "transformations": [{"name": "cyclic", "range": [0, 360]}]},
                {"axis_name": "step", "transformations": [{"name": "type_change", "type": "int"}]},
                {"axis_name": "number", "transformations": [{"name": "type_change", "type": "int"}]},
                {"axis_name": "levelist", "transformations": [{"name": "type_change", "type": "int"}]},
            ],
            "compressed_axes_config": [
                "longitude",
                "latitude",
                "levtype",
                "levelist",
                "step",
                "date",
                "domain",
                "expver",
                "param",
                "class",
                "stream",
                "type",
                "number",
            ],
            "pre_path": {
                "class": "od",
                "expver": "0001",
                "levtype": "pl",
                "stream": "enfo",
                "type": "pf",
                "domain": "g",
                "date": self.date,
                "time": "0000",
            },
        }

        conf = Conflator(app_name="polytope_mars", model=PolytopeMarsConfig).load()
        self.cf = conf.model_dump()
        self.cf["options"] = self.options

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_trajectory(self):
        result = PolytopeMars(self.cf).extract(self.request)
        decoder = Covjsonkit().decode(result)
        decoder.to_xarray()
        assert True

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_trajectory_latlon_correct_inflation(self):
        self.request["feature"]["inflation"] = [0.1, 0.2]
        result = PolytopeMars(self.cf).extract(self.request)
        decoder = Covjsonkit().decode(result)
        decoder.to_xarray()
        assert True

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_trajectory_latlon_incorrect_inflates(self):
        with pytest.raises(ValueError):
            self.request["feature"]["inflation"] = [0.1, 0.2, 0.3]
            result = PolytopeMars(self.cf).extract(self.request)
            decoder = Covjsonkit().decode(result)
            decoder.to_xarray()

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_trajectory_lonlat(self):
        self.request["feature"]["axes"] = ["longitude", "latitude"]
        result = PolytopeMars(self.cf).extract(self.request)
        decoder = Covjsonkit().decode(result)
        decoder.to_xarray()
        assert True

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_trajectory_3d_levelist(self):
        self.request["feature"]["axes"] = ["latitude", "longitude", "levelist"]
        self.request["feature"]["points"] = [[-1, -1, 1], [0, 0, 2], [1, 1, 10]]
        self.request["feature"]["inflate"] = "round"
        del self.request["levelist"]
        result = PolytopeMars(self.cf).extract(self.request)
        decoder = Covjsonkit().decode(result)
        decoder.to_xarray()
        assert True

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_trajectory_3d_step(self):
        self.request["feature"]["axes"] = ["latitude", "longitude", "step"]
        self.request["feature"]["points"] = [[-1, -1, 1], [0, 0, 2], [1, 1, 10]]
        self.request["feature"]["inflate"] = "round"
        del self.request["step"]
        result = PolytopeMars(self.cf).extract(self.request)
        decoder = Covjsonkit().decode(result)
        decoder.to_xarray()
        assert True

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_trajectory_4d(self):
        request_copy = copy.deepcopy(self.request)
        self.request["feature"]["axes"] = ["latitude", "longitude", "levelist", "step"]
        self.request["feature"]["points"] = [[-1, -1, 1, 1], [0, 0, 2, 2], [1, 1, 10, 3]]
        self.request["feature"]["inflate"] = "box"
        del self.request["levelist"]
        del self.request["step"]
        result1 = PolytopeMars(self.cf).extract(self.request)

        request_copy["feature"]["axes"] = ["latitude", "longitude", "step", "levelist"]
        request_copy["feature"]["points"] = [[-1, -1, 1, 1], [0, 0, 2, 2], [1, 1, 3, 10]]
        request_copy["feature"]["inflate"] = "box"
        del request_copy["levelist"]
        del request_copy["step"]
        result2 = PolytopeMars(self.cf).extract(request_copy)

        assert result1 == result2

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_trajectory_4d_mix_axes(self):
        request_copy = copy.deepcopy(self.request)
        self.request["feature"]["axes"] = ["latitude", "longitude", "levelist", "step"]
        self.request["feature"]["points"] = [[-1, -1, 1, 1], [0, 0, 2, 2], [1, 1, 10, 3]]
        self.request["feature"]["inflate"] = "box"
        del self.request["levelist"]
        del self.request["step"]
        result1 = PolytopeMars(self.cf).extract(self.request)

        request_copy["feature"]["axes"] = ["longitude", "step", "latitude", "levelist"]
        request_copy["feature"]["points"] = [[-1, 1, -1, 1], [0, 2, 0, 2], [1, 3, 1, 10]]
        request_copy["feature"]["inflate"] = "box"
        del request_copy["levelist"]
        del request_copy["step"]
        result2 = PolytopeMars(self.cf).extract(request_copy)
        assert result1 == result2

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_trajectory_1_axes(self):
        self.request["feature"]["axes"] = ["latitude"]
        self.request["feature"]["points"] = [[-1], [0], [1]]
        with pytest.raises(KeyError):
            PolytopeMars(self.cf).extract(self.request)

    def test_trajectory_no_inflation(self):
        del self.request["feature"]["inflation"]
        with pytest.raises(KeyError):
            PolytopeMars(self.cf).extract(self.request)

    def test_trajectory_levelist_multiple(self):
        self.request["levelist"] = "500/700"
        result = PolytopeMars(self.cf).extract(self.request)
        decoder = Covjsonkit().decode(result)
        decoder.to_xarray()
        assert True



import logging

from polytope_feature import shapes

from ..feature import Feature
from ..utils.areas import field_area


class TimeSeries(Feature):
    def __init__(self, feature_config, client_config):
        assert feature_config.pop("type") == "timeseries"
        # self.start_step = config.pop("start", None)
        # self.end_step = config.pop("end", None)
        self.axes = feature_config.pop("axes", [])
        self.time_axis = feature_config.pop("time_axis", [])

        self.max_size = client_config.polygonrules.max_area

        if self.axes != []:
            if not isinstance(self.axes, list):
                self.axes = ["latitude", "longitude"]
            if self.axes == ["step"] or self.axes == ["date"]:
                self.axes = ["latitude", "longitude"]
        else:
            self.axes = ["latitude", "longitude"]

        self.points = feature_config.pop("points", [])

        if "range" in feature_config:
            feature_config.pop("range")

        assert len(feature_config) == 0, f"Unexpected keys in config: {feature_config.keys()}"

    def get_shapes(self):
        # Time-series is a squashed box from start_step to start_end for each point  # noqa: E501
        return [
            shapes.Union(
                [self.axes[0], self.axes[1]],
                *[
                    shapes.Point(
                        [self.axes[0], self.axes[1]],
                        [[p[0], p[1]]],
                        method="nearest",  # noqa: E501
                    )
                    for p in self.points
                ],
            ),
            # shapes.Span("step", self.start_step, self.end_step),
        ]

    def incompatible_keys(self):
        return ["levellist"]

    def coverage_type(self):
        return "PointSeries"

    def name(self):
        return "Time Series"

    def required_keys(self):
        return ["type", "points", "time_axis"]

    def required_axes(self):
        return ["latitude", "longitude"]

    def parse(self, request, feature_config):
        logging.debug("Feature config: %s", feature_config)
        # if isinstance(feature_config["time_axis"], list):
        #    if "step" not in feature_config["time_axis"] and "date" not in feature_config["time_axis"]:
        #        raise ValueError("Timeseries axes must be step or date")
        if feature_config["time_axis"] != "step" and feature_config["time_axis"] != "date":  # noqa: E501
            raise ValueError("Timeseries axes must be step or date")

        area = field_area(request, len(feature_config["points"]))

        if area > self.max_size:
            raise ValueError(
                f"Number of coordinates*fields for timeseries {area} exceeds total number allowed, please reduce the number of coordinates or fields requested"  # noqa: E501
            )

        if isinstance(feature_config["time_axis"], list):
            if "step" in feature_config["time_axis"]:
                time_axis = "step"
                feature_config["time_axis"].remove("step")
            if "date" in feature_config["time_axis"]:
                time_axis = "date"
                feature_config["time_axis"].remove("date")
        else:
            time_axis = feature_config["time_axis"]

        if len(feature_config["points"][0]) != 2:
            raise ValueError("Timeseries must have only two values in points")
        if time_axis in request and "range" in feature_config:
            raise ValueError("Timeseries time_axis is overspecified in request")
        if time_axis not in request and "range" not in feature_config:  # noqa: E501
            raise ValueError("Timeseries time_axis is underspecified in request")

        if "range" in feature_config:
            if feature_config["range"]["start"] < 0:
                raise ValueError("Timeseries range start must be greater than 0")
            if isinstance(feature_config["range"], dict):
                time_range = f"{feature_config['range']['start']}/to/{feature_config['range']['end']}"
                request[time_axis] = time_range  # noqa: E501
                if "interval" in feature_config["range"]:
                    request[time_axis] += f"/by/{feature_config['range']['interval']}"
        logging.debug("After parse request: %s", request)

        return request



import re
from datetime import datetime, timedelta

import pandas as pd


def days_between_dates(date1, date2):
    """
    Calculate the number of days between two dates in the format YYYYMMDD.

    :param date1: The first date in the format YYYYMMDD
    :param date2: The second date in the format YYYYMMDD
    :return: The number of days between the two dates
    """
    date_format = "%Y%m%d"
    d1 = datetime.strptime(date1, date_format)
    d2 = datetime.strptime(date2, date_format)
    delta = d2 - d1
    return abs(delta.days)


def hours_between_times(time1, time2):
    """
    Calculate the number of hours between two times in the format HHMM.

    :param time1: The first time in the format HHMM
    :param time2: The second time in the format HHMM
    :return: The number of hours between the two times
    """
    time_format = "%H%M"
    t1 = datetime.strptime(time1, time_format)
    t2 = datetime.strptime(time2, time_format)
    delta = t2 - t1
    return abs(delta.total_seconds() / 3600)


def convert_timestamp(timestamp):
    # Ensure the input is a string
    timestamp = str(timestamp)

    # Pad the timestamp with leading zeros if necessary
    timestamp = timestamp.zfill(4)

    # Insert colons to format as HH:MM:SS
    formatted_timestamp = f"{timestamp[:2]}:{timestamp[2:]}:00"

    return formatted_timestamp


def find_step_intervals(step_start: str, step_end: str, step_freq: str):
    def format_subhourly_step_to_mars(step: pd.Timedelta):
        total_hours = int(step.total_seconds() // 3600)
        minutes = int((step.total_seconds() % 3600) // 60)
        return f"{total_hours}h{minutes}m"

    def format_step_str_to_pd(step: str):
        return re.sub(r"m$", "min", step)

    # if we have normal digit steps only, treat like before
    if step_start.isdigit() and step_end.isdigit() and step_freq.isdigit():
        return list(range(int(step_start), int(step_end), int(step_freq)))

    step_start_pd_format = format_step_str_to_pd(step_start)
    step_end_pd_format = format_step_str_to_pd(step_end)
    step_freq_pd_format = format_step_str_to_pd(step_freq)

    step_values = pd.timedelta_range(start=step_start_pd_format, end=step_end_pd_format, freq=step_freq_pd_format)
    return [format_subhourly_step_to_mars(val) for val in step_values]


def from_range_to_list_num(num_range):
    """
    Convert a range in the format integer/to/integer to a list of numbers seperated by /.

    :param range: A string representing the range in the format integer/to/integer
    :return: A list of dates in the format integer/integer/integer
    """
    if "/to/" in num_range:
        start, end = num_range.split("/to/")
        start = int(start)
        end = int(end)
        if start > end:
            raise ValueError("Start of range must be less than or equal to end of range.")
        return [str(i) for i in range(start, end + 1)]
    else:
        return num_range


def from_range_to_list_date(date_range):
    """
    Convert a date range in the format YYYYMMDD/to/YYYYMMDD to a list of dates.

    :param date_range: A string representing the date range in the format YYYYMMDD/to/YYYYMMDD
    :return: A list of dates in the format YYYYMMDD/YYYYMMDD
    """
    if "/to/" in date_range:
        start_date, end_date = date_range.split("/to/")
        start_date = datetime.strptime(start_date, "%Y%m%d")
        end_date = datetime.strptime(end_date, "%Y%m%d")
        date_list = []
        current_date = start_date
        while current_date <= end_date:
            date_list.append(current_date.strftime("%Y%m%d"))
            current_date += timedelta(days=1)
        if not date_list:
            raise ValueError("The date range does not include any valid dates.")
        if len(date_list) == 1:
            return date_list[0]
        date_list = "/".join(date_list)
    else:
        date_list = date_range
    return date_list



from polytope_mars.utils.datetimes import from_range_to_list_date


def test_from_range_to_list_date_spans_months():
    date_range = "20220130/to/20220202"
    expected = "20220130/20220131/20220201/20220202"
    assert from_range_to_list_date(date_range) == expected



from datetime import datetime, timedelta

import pytest
from conflator import Conflator
from covjsonkit.api import Covjsonkit

from polytope_mars.api import PolytopeMars
from polytope_mars.config import PolytopeMarsConfig

# If using a local FDB need to set GRIBJUMP_CONFIG_FILE and DYLD_LIBRARY_PATH


class TestFeatureFactory:
    def setup_method(self):

        today = datetime.today()
        yesterday = today - timedelta(days=1)
        self.date = yesterday.strftime("%Y%m%d")

        self.request = {
            "class": "od",
            "stream": "enfo",
            "type": "pf",
            "date": self.date,
            "time": "0000",
            "levtype": "sfc",
            "expver": "0001",
            "domain": "g",
            "param": "164/166/167/169",
            "number": "1/to/2",
            "feature": {
                "type": "timeseries",
                "points": [[-9.10, 38.78]],
                "axes": "step",
                "range": {"start": 0, "end": 3},
            },
        }

        self.options = {
            "axis_config": [
                {
                    "axis_name": "date",
                    "transformations": [{"name": "merge", "other_axis": "time", "linkers": ["T", "00"]}],
                },
                {
                    "axis_name": "values",
                    "transformations": [
                        {
                            "name": "mapper",
                            "type": "octahedral",
                            "resolution": 1280,
                            "axes": ["latitude", "longitude"],
                        }
                    ],
                },
                {"axis_name": "levelist", "transformations": [{"name": "type_change", "type": "str"}]},
                {"axis_name": "latitude", "transformations": [{"name": "reverse", "is_reverse": True}]},
                {"axis_name": "longitude", "transformations": [{"name": "cyclic", "range": [0, 360]}]},
                {"axis_name": "step", "transformations": [{"name": "type_change", "type": "int"}]},
                {"axis_name": "number", "transformations": [{"name": "type_change", "type": "int"}]},
            ],
            "compressed_axes_config": [
                "longitude",
                "latitude",
                "levtype",
                "step",
                "date",
                "domain",
                "expver",
                "param",
                "class",
                "stream",
                "type",
                "number",
            ],
            "pre_path": {
                "class": "od",
                "expver": "0001",
                "levtype": "sfc",
                "stream": "enfo",
                "type": "pf",
                "domain": "g",
            },
        }

        conf = Conflator(app_name="polytope_mars", model=PolytopeMarsConfig).load()
        self.cf = conf.model_dump()
        self.cf["options"] = self.options

    # @pytest.mark.skip(reason="Gribjump not set up for ci actions yet")
    def test_timeseries(self):
        result = PolytopeMars(self.cf).extract(self.request)
        decoder = Covjsonkit().decode(result)
        decoder.to_xarray()
        assert True

    def test_timeseries_step_in_both(self):
        self.request["step"] = "0/to/3"
        with pytest.raises(ValueError):
            result = PolytopeMars(self.cf).extract(self.request)
            decoder = Covjsonkit().decode(result)
            decoder.to_xarray()

    def test_timeseries_step_in_request(self):
        self.request["step"] = "0/to/3"
        del self.request["feature"]["range"]
        result = PolytopeMars(self.cf).extract(self.request)
        decoder = Covjsonkit().decode(result)
        decoder.to_xarray()
        assert True

    def test_timeseries_no_axes(self):
        with pytest.raises(KeyError):
            del self.request["feature"]["axes"]
            PolytopeMars(self.cf).extract(self.request)



import datetime
import json
import logging
import time
from typing import List

import pandas as pd
import pygribjump as gj
from conflator import Conflator
from covjsonkit.api import Covjsonkit
from covjsonkit.param_db import get_param_ids
from covjsonkit.utils import merge_coverage_collections
from polytope_feature import shapes
from polytope_feature.polytope import Polytope, Request

from .config import PolytopeMarsConfig
from .features.boundingbox import BoundingBox
from .features.circle import Circle
from .features.frame import Frame
from .features.path import Path
from .features.polygon import Polygons
from .features.position import Position
from .features.shpfile import Shapefile
from .features.timeseries import TimeSeries
from .features.verticalprofile import VerticalProfile
from .utils.datetimes import (
    convert_timestamp,
    find_step_intervals,
    from_range_to_list_date,
    from_range_to_list_num,
)

features = {
    "timeseries": TimeSeries,
    "verticalprofile": VerticalProfile,
    "boundingbox": BoundingBox,
    "frame": Frame,
    "trajectory": Path,
    "shapefile": Shapefile,
    "polygon": Polygons,
    "circle": Circle,
    "position": Position,
}


class PolytopeMars:
    def __init__(self, config=None, log_context=None):
        # Initialise polytope-mars configuration
        self.log_context = log_context
        self.id = log_context["id"] if log_context else "-1"

        # If no config check default locations
        if config is None:
            self.conf = Conflator(app_name="polytope_mars", model=PolytopeMarsConfig).load()  # noqa: E501
            logging.debug(f"{self.id}: Config loaded from file: {self.conf}")  # noqa: E501
        # else initialise with provided config
        else:
            self.conf = PolytopeMarsConfig.model_validate(config)
            logging.debug(f"{self.id}: Config loaded from dictionary: {self.conf}")  # noqa: E501

        self.coverage = {}
        self.split_request = False

    def extract(self, request):
        # request expected in JSON or dict
        if not isinstance(request, dict):
            try:
                request = json.loads(request)
            except ValueError:
                raise ValueError("Request not in JSON format or python dictionary")  # noqa: E501

        # expect a "feature" key in the request
        try:
            feature_config = request.pop("feature")
            feature_config_copy = feature_config.copy()
        except KeyError:
            raise KeyError("Request does not contain a 'feature' keyword")

        try:
            format = request.pop("format")
            if format != "covjson":
                raise ValueError("Only covjson format is currently supported")
        except KeyError:
            pass

        # get feature type
        try:
            feature_type = feature_config["type"]
        except KeyError:
            raise KeyError("The 'feature' does not contain a 'type' keyword")

        if feature_type == "timeseries":
            try:
                if "time_axis" in feature_config:
                    timeseries_type = feature_config["time_axis"]
                if "axes" in feature_config:
                    if feature_config["axes"] == "step":
                        timeseries_type = "step"
                        del feature_config["axes"]
                        del feature_config_copy["axes"]
                        feature_config["time_axis"] = "step"
                        feature_config_copy["time_axis"] = "step"
                    elif "step" in feature_config["axes"]:
                        raise ValueError(
                            "Step axis not supported in 'axes' keyword, must be in 'time_axis'"
                        )  # noqa: E501
                    elif "date" in feature_config["axes"]:
                        raise ValueError(
                            "Date axis not supported in 'axes' keyword, must be in 'time_axis'"
                        )  # noqa: E501

            except KeyError:
                raise KeyError("The timeseries feature requires a 'time_axis' keyword")  # noqa: E501
        else:
            timeseries_type = None  # noqa: F841

        feature = self._feature_factory(feature_type, feature_config, self.conf)  # noqa: E501

        feature.validate(request, feature_config_copy)

        logging.debug("Unparsed request: %s", request)
        logging.debug("Feature dictionary: %s", feature_config_copy)

        request = feature.parse(request, feature_config_copy)

        self.split_request = feature.split_request()

        logging.debug("Self split: %s", self.split_request)
        logging.debug("Parsed request: %s", request)

        if self.split_request:
            # If the request is split, we need to handle it differently
            dates = from_range_to_list_date(request["date"])
            for date in dates.split("/"):
                if "number" in request:
                    numbers = from_range_to_list_num(request["number"])
                    if len(numbers) > 10:
                        for number in from_range_to_list_num(request["number"]):
                            copied_request = request.copy()
                            copied_request["date"] = date
                            copied_request["number"] = number
                            coverage = self.retrieve_data(copied_request, feature_type, feature)  # noqa: E501
                            self.coverage = merge_coverage_collections(self.coverage, coverage)  # noqa: E501
                    else:
                        copied_request = request.copy()
                        copied_request["date"] = date
                        coverage = self.retrieve_data(copied_request, feature_type, feature)
                        self.coverage = merge_coverage_collections(self.coverage, coverage)
                else:
                    copied_request = request.copy()
                    copied_request["date"] = date
                    coverage = self.retrieve_data(copied_request, feature_type, feature)
                    self.coverage = merge_coverage_collections(self.coverage, coverage)

        else:
            self.coverage = self.retrieve_data(request, feature_type, feature)  # noqa: E501

        return self.coverage

    def _create_base_shapes(self, request: dict, feature_type) -> List[shapes.Shape]:
        base_shapes = []

        if (
            "dataset" in request
            and request["dataset"] == "climate-dt"  # noqa: W503
            and (feature_type == "timeseries" or feature_type == "polygon")  # noqa: W503
        ) or (request["class"] == "ng" and (feature_type == "timeseries" or feature_type == "polygon")):
            for k, v in request.items():
                split = str(v).split("/")

                if k == "param":
                    try:
                        int(split[0])
                    except:  # noqa: E722
                        new_split = []
                        for s in split:
                            new_split.append(get_param_ids(self.conf.coverageconfig)[s])  # noqa: E501
                        split = new_split

                # ALL -> All
                if len(split) == 1 and split[0] == "ALL":
                    base_shapes.append(shapes.All(k))

                # Single value -> Select
                elif len(split) == 1:
                    if k == "date":
                        split[0] = pd.Timestamp(split[0])
                    if k == "time":
                        split[0] = convert_timestamp(split[0])
                    base_shapes.append(shapes.Select(k, split))

                # Range a/to/b, "by" not supported -> Span
                elif len(split) == 3 and split[1] == "to":
                    # if date then only get time of dates in span not
                    # all in times within date
                    if k == "date":
                        start = pd.Timestamp(split[0])
                        end = pd.Timestamp(split[2])
                        base_shapes.append(shapes.Span(k, lower=start, upper=end))
                    elif k == "time":
                        start = convert_timestamp(split[0])
                        end = convert_timestamp(split[2])
                        base_shapes.append(shapes.Span(k, lower=start, upper=end))
                    else:
                        base_shapes.append(shapes.Span(k, lower=split[0], upper=split[2]))  # noqa: E501

                elif "by" in split:

                    if split[-1] == "1":
                        if k == "date":
                            start = pd.Timestamp(split[0])
                            end = pd.Timestamp(split[2])
                            base_shapes.append(shapes.Span(k, lower=start, upper=end))
                        elif k == "time":
                            start = convert_timestamp(split[0])
                            end = convert_timestamp(split[2])
                            base_shapes.append(shapes.Span(k, lower=start, upper=end))
                        else:
                            base_shapes.append(shapes.Span(k, lower=split[0], upper=split[2]))  # noqa: E501
                    else:
                        if k == "date":
                            start = pd.Timestamp(split[0])
                            end = pd.Timestamp(split[2])
                            timestamps = pd.date_range(start=start, end=end, freq=f"{split[-1]}D")
                            base_shapes.append(shapes.Select(k, timestamps.tolist()))
                        elif k == "time":
                            start = convert_timestamp(split[0])
                            end = convert_timestamp(split[2])
                            times = pd.date_range(start=start, end=end, freq=f"{split[-1]}H")
                            base_shapes.append(shapes.Select(k, times.strftime("%H:%M:%S").tolist()))
                            # base_shapes.append(shapes.Span(k, lower=start, upper=end))
                            # base_shapes.append(shapes.Span(k, lower=start, upper=end))
                        # raise ValueError("Ranges with step-size specified with 'by' keyword is not supported")  # noqa: E501

                # List of individual values -> Union of Selects
                else:
                    if k == "date":
                        dates = []
                        for s in split:
                            dates.append(pd.Timestamp(s))
                        split = dates
                    if k == "time":
                        times = []
                        for s in split:
                            times.append(convert_timestamp(s))
                        split = times
                    base_shapes.append(shapes.Select(k, split))
        else:
            time = request.pop("time").replace(":", "")
            time = time.split("/")
            if "to" in time:
                start = convert_timestamp(time[0])
                end = convert_timestamp(time[2])
                if "by" in time:
                    times = pd.date_range(start=start, end=end, freq=f"{time[-1]}H")
                else:
                    times = pd.date_range(start=start, end=end, freq="1H")
                time = times.strftime("%H:%M:%S").tolist()

            for k, v in request.items():
                split = str(v).split("/")

                if k == "param":
                    try:
                        int(split[0])
                    except:  # noqa: E722
                        new_split = []
                        for s in split:
                            new_split.append(get_param_ids(self.conf.coverageconfig)[s])  # noqa: E501
                        split = new_split

                # ALL -> All
                if len(split) == 1 and split[0] == "ALL":
                    base_shapes.append(shapes.All(k))

                # Single value -> Select
                elif len(split) == 1:
                    if k == "date":
                        if int(split[0]) < 0:
                            split[0] = str(
                                (
                                    datetime.datetime.now() + datetime.timedelta(days=int(split[0]))
                                ).strftime(  # noqa: E501
                                    "%Y%m%d"
                                )  # noqa: E501
                            )
                        new_split = []
                        for t in time:
                            new_split.append(pd.Timestamp(split[0] + "T" + t))
                        split = new_split
                    base_shapes.append(shapes.Select(k, split))

                # Range a/to/b, "by" not supported -> Span
                elif len(split) == 3 and split[1] == "to":
                    # if date then only get time of dates in span not
                    # all in times within date
                    if k == "date":
                        start = pd.Timestamp(split[0] + "T" + time[0])
                        end = pd.Timestamp(split[2] + "T" + time[-1])
                        dates = []
                        for s in pd.date_range(start, end):
                            for t in time:
                                dates.append(pd.Timestamp(s.strftime("%Y%m%d") + "T" + t))
                        base_shapes.append(shapes.Select(k, dates))
                    else:
                        base_shapes.append(shapes.Span(k, lower=split[0], upper=split[2]))  # noqa: E501

                elif "by" in split:
                    if split[-1] == "1":
                        if k == "date":
                            start = pd.Timestamp(split[0] + "T" + time[0])
                            end = pd.Timestamp(split[2] + "T" + time[-1])
                            dates = []
                            for s in pd.date_range(start, end):
                                for t in time:
                                    dates.append(pd.Timestamp(s.strftime("%Y%m%d") + "T" + t))
                            base_shapes.append(shapes.Select(k, dates))
                        else:
                            base_shapes.append(shapes.Span(k, lower=split[0], upper=split[2]))
                    else:
                        if k == "date":
                            start = pd.Timestamp(split[0] + "T" + time[0])
                            end = pd.Timestamp(split[2] + "T" + time[-1])
                            dates = []
                            for s in pd.date_range(start, end, freq=f"{split[-1]}D"):
                                for t in time:
                                    dates.append(pd.Timestamp(s.strftime("%Y%m%d") + "T" + t))
                            base_shapes.append(shapes.Select(k, dates))
                        elif k == "step":
                            steps = find_step_intervals(split[0], split[2], split[-1])
                            base_shapes.append(shapes.Select(k, steps))
                        else:
                            expansion = list(range(int(split[0]), int(split[2]), int(split[-1])))
                            base_shapes.append(shapes.Select(k, expansion))

                # List of individual values -> Union of Selects
                else:
                    if k == "date":
                        dates = []
                        for s in split:
                            for t in time:
                                dates.append(pd.Timestamp(s + "T" + t))
                        split = dates
                    base_shapes.append(shapes.Select(k, split))

        return base_shapes

    def _feature_factory(self, feature_name, feature_config, config=None):
        feature_class = features.get(feature_name)
        if feature_class:
            return feature_class(feature_config, config)
        else:
            raise NotImplementedError(f"Feature '{feature_name}' not found")

    def retrieve_data(self, request, feature_type, feature):
        """
        Retrieves data from the Polytope engine based on the request and feature type.
        This method sets up the Polytope engine, prepares the request, and encodes the
        result into a Covjson format.

        :param request: The request dictionary containing parameters for data retrieval.
        :param feature_type: The type of feature being requested (e.g., 'timeseries', 'polygon').
        :param feature: The feature object that contains the logic for data retrieval.
        :return: The coverage data in Covjson format.
        """
        shapes = self._create_base_shapes(request, feature_type)

        shapes.extend(feature.get_shapes())

        preq = Request(*shapes)

        start = time.time()
        logging.info(f"{self.id}: Gribjump/setup time start: {start}")  # noqa: E501

        if self.conf.datacube.type == "gribjump":
            fdbdatacube = gj.GribJump()
        else:
            raise NotImplementedError(f"Datacube type '{self.conf.datacube.type}' not found")  # noqa: E501

        logging.debug(f"Send log_context to polytope: {self.log_context}")
        self.api = Polytope(
            datacube=fdbdatacube,
            options=self.conf.options.model_dump(),
            context=self.log_context,
        )

        end = time.time()
        delta = end - start
        logging.debug(f"{self.id}: Gribjump/setup time end: {end}")  # noqa: E501
        logging.info(f"{self.id}: Gribjump/setup time taken: {delta}")  # noqa: E501

        logging.debug(f"{self.id}: The request we give polytope from polytope-mars is: {preq}")  # noqa: E501
        start = time.time()
        logging.info(f"{self.id}: Polytope time start: {start}")  # noqa: E501

        result = self.api.retrieve(preq)

        end = time.time()
        delta = end - start
        logging.debug(f"{self.id}: Polytope time end: {end}")  # noqa: E501
        logging.info(f"{self.id}: Polytope time taken: {delta}")  # noqa: E501
        start = time.time()
        logging.info(f"{self.id}: Covjson time start: {start}")  # noqa: E501
        encoder = Covjsonkit(self.conf.coverageconfig.model_dump()).encode(
            "CoverageCollection", feature_type
        )  # noqa: E501

        if "dataset" in request:
            if request["dataset"] == "climate-dt" and (feature_type == "timeseries" or feature_type == "polygon"):
                coverage = encoder.from_polytope_step(result)
            else:
                coverage = encoder.from_polytope(result)
        elif request["class"] == "ng":  # noqa: E501
            if feature_type == "timeseries" or feature_type == "polygon":
                coverage = encoder.from_polytope_step(result)
            else:
                coverage = encoder.from_polytope(result)
        else:
            coverage = encoder.from_polytope(result)

        end = time.time()
        delta = end - start
        logging.debug(f"{self.id}: Covjsonkit time end: {end}")  # noqa: E501
        logging.info(f"{self.id}: Covjsonkit time taken: {delta}")  # noqa: E501

        return coverage



from conflator import ConfigModel
from polytope_feature.options import Config


class DatacubeConfig(ConfigModel):
    type: str = "gribjump"
    config: str = "config.yaml"
    uri: str = "http://localhost:8000"


class CovjsonKitConfig(ConfigModel):
    param_db: str = "ecmwf"


class PolygonRulesConfig(ConfigModel):
    # Max points is the max number of points in all polygons requested allowed
    max_points: int = 1000
    # Max area is the max area of all polygons requested that is allowed.
    # Area is calculated in kilometers squared
    max_area: float = float("inf")


class PolytopeMarsConfig(ConfigModel):
    datacube: DatacubeConfig = DatacubeConfig()
    options: Config = Config()
    coverageconfig: CovjsonKitConfig = CovjsonKitConfig()
    polygonrules: PolygonRulesConfig = PolygonRulesConfig()



import io
import re

from setuptools import find_packages, setup

__version__ = re.search(
    r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
    io.open("polytope_mars/version.py", encoding="utf_8_sig").read(),
).group(1)


with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="polytope_mars",
    version=__version__,
    description="High level meteorological feature extraction interface to Polytope",  # noqa: E501
    long_description="",
    url="https://github.com/ecmwf/polytope-mars",
    author="ECMWF",
    author_email="James.Hawkes@ecmwf.int, Adam.Warde@ecmwf.int, Mathilde.Leuridan@ecmwf.int",  # noqa: E501
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    install_requires=requirements,
)



import math

import numpy as np
from geographiclib.geodesic import Geodesic
from geographiclib.polygonarea import PolygonArea
from shapely.geometry import LineString, Polygon
from shapely.ops import split

from .datetimes import days_between_dates, hours_between_times


def haversine_distance(lat1, lon1, lat2, lon2):
    """
    Calculate the great-circle distance between two points on the Earth's surface.

    :param lat1: Latitude of the first point in degrees
    :param lon1: Longitude of the first point in degrees
    :param lat2: Latitude of the second point in degrees
    :param lon2: Longitude of the second point in degrees
    :return: Distance between the two points in kilometers
    """
    R = 6371  # Radius of the Earth in kilometers
    phi1 = math.radians(lat1)
    phi2 = math.radians(lat2)
    delta_phi = math.radians(lat2 - lat1)
    delta_lambda = math.radians(lon2 - lon1)

    a = math.sin(delta_phi / 2) ** 2 + math.cos(phi1) * math.cos(phi2) * math.sin(delta_lambda / 2) ** 2
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))

    return R * c


def get_circle_area(radius_km):
    """
    Calculate the area of a circle on the surface of the Earth in square kilometers.

    :param radius_km: Radius of the circle in kilometers
    :return: Area of the circle in square kilometers
    """
    return math.pi * (radius_km**2)


def get_circle_area_from_coords(lat_center, lon_center, lat_point, lon_point):
    """
    Calculate the area of a circle on the surface of the Earth in square kilometers
    given the coordinates of the center and a point on the circumference.

    :param lat_center: Latitude of the circle's center in degrees
    :param lon_center: Longitude of the circle's center in degrees
    :param lat_point: Latitude of a point on the circumference in degrees
    :param lon_point: Longitude of a point on the circumference in degrees
    :return: Area of the circle in square kilometers
    """
    radius_km = haversine_distance(lat_center, lon_center, lat_point, lon_point)
    return get_circle_area(radius_km)


def split_polygon(polygon):
    minx, miny, maxx, maxy = polygon.bounds

    # Determine all multiples of 90 degrees within the longitude range
    # We extend the range slightly to include exact multiples if
    # they coincide with minx or maxx
    start_lon = int(np.floor(minx / 90.0)) * 90
    end_lon = int(np.ceil(maxx / 90.0)) * 90
    split_longitudes = list(range(start_lon, end_lon + 1, 90))

    pieces = [polygon]

    # Split the polygon using the specified meridians
    for lon in split_longitudes:
        new_pieces = []
        splitter = LineString([(lon, -90), (lon, 90)])
        for poly in pieces:
            if not poly.is_empty:
                result = split(poly, splitter)
                for geom in result.geoms:
                    if geom.geom_type in ["Polygon", "MultiPolygon"]:
                        new_pieces.append(geom)
            else:
                continue
        pieces = new_pieces

    return pieces


def get_area_piece(piece):
    geod = Geodesic.WGS84
    polygon = PolygonArea(geod, False)

    if piece.geom_type == "Polygon":
        coords = list(piece.exterior.coords)
    elif piece.geom_type == "MultiPolygon":
        area = 0.0
        for poly in piece:
            area += get_area_piece(poly)
        return area
    else:
        return 0.0

    for lon, lat in coords:
        polygon.AddPoint(lat, lon)

    num, perimeter, area = polygon.Compute(False, True)

    area = abs(area)

    return area  # area in square meters


def get_polygon_area(points):
    polygon = Polygon([(lon, lat) for lat, lon in points])
    pieces = split_polygon(polygon)
    total_area = 0.0
    for piece in pieces:
        area = get_area_piece(piece)
        total_area += area
    return total_area / 1e6  # Convert area from square meters to square kilometers  # noqa: E501


def get_boundingbox_area(points):
    # Convert points to a Shapely Polygon
    min_lon, min_lat = points[0][:2]
    max_lon, max_lat = points[1][:2]
    if min_lon + max_lon == 0:
        min_lon += 0.1

    # Define the polygon coordinates
    polygon_coords = [
        (min_lat, min_lon),  # Bottom-left corner
        (min_lat, max_lon),  # Bottom-right corner
        (max_lat, max_lon),  # Top-right corner
        (max_lat, min_lon),  # Top-left corner
        (
            min_lat,
            min_lon,
        ),  # Closing the polygon by returning to the bottom-left corner
    ]
    polygon = Polygon(polygon_coords)
    pieces = split_polygon(polygon)
    total_area = 0.0
    for piece in pieces:
        area = get_area_piece(piece)
        total_area += area
    return total_area / 1e6  # Convert area from square meters to square kilometers  # noqa: E501


def field_area(request, area):
    """
    Calculate the area of a request based on the number of fields and the area of the feature.
    :param request: The request dictionary containing fields and feature dictionary.
    :param area: The area of the feature in square kilometers.
    :return: The total area of the request in square kilometers.
    """

    step_len = 1
    number_len = 1
    levelist_len = 1

    # Check if the request contains a range for the feature instead of a step
    if "feature" in request:
        if "range" in request["feature"]:
            if "start" in request["feature"]["range"] and "end" in request["feature"]["range"]:
                step_len = request["feature"]["range"]["end"] - request["feature"]["range"]["start"] + 1
            elif "start" in request["feature"]["range"]:
                step_len = 1
            elif "end" in request["feature"]["range"]:
                step_len = 1

    if "step" in request:
        step = request["step"].split("/")
        if "to" in step:
            step_len = len(range(int(step[0]), int(step[2]) + 1))
        else:
            step_len = len(step)

    if "number" in request:
        number = request["number"].split("/")
        if "to" in number:
            number_len = len(range(int(number[0]), int(number[2]) + 1))
        else:
            number_len = len(number)

    if "levelist" in request:
        levelist = request["levelist"].split("/")
        if "to" in levelist:
            levelist_len = len(range(int(levelist[0]), int(levelist[2]) + 1))
        else:
            levelist_len = len(levelist)

    date = request["date"].split("/")
    time = request["time"].split("/")
    param = request["param"].split("/")

    if "to" in date:
        date_len = days_between_dates(date[0], date[2])
    else:
        date_len = len(date)

    if date_len == 0:
        date_len = 1

    if "to" in time:
        time_len = hours_between_times(time[0], time[2])
    else:
        time_len = len(time)

    param_len = len(param)

    shape_area = area

    return param_len * step_len * number_len * time_len * date_len * levelist_len * shape_area


def request_cost(request):
    """
    Calculate the cost of a request based on the area and the number of fields.
    Note this is only a heuristic and does not take into account the actual cost of the request.

    :param request: The request dictionary containing fields and feature dictionary.
    :return: The cost of the request.
    """
    if request["feature"]["type"] == "boundingbox":
        area = get_boundingbox_area(request["feature"]["points"])
    elif request["feature"]["type"] == "polygon":
        area = get_polygon_area(request["feature"]["shape"])
    else:
        area = len(request["feature"]["points"])
    field_area_value = field_area(request, area)

    if "dataset" in request:
        field_area_value *= 2

    return field_area_value
